-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}

M.values = {}
local lastValues = {}
local smoothValuesExp = {}
M.disabledState = {}

MFM = require("MFM")

local smoothers = {
    wheelspeed = newExponentialSmoothing(10),
    gear_A = newExponentialSmoothing(10),
    gear_M = newExponentialSmoothing(10),
    rpm = newExponentialSmoothing(10),
    lights = newExponentialSmoothing(10),
    fuel = newExponentialSmoothing(10),
    oiltemp = newExponentialSmoothing(10),
    watertemp = newExponentialSmoothing(10),
    turnsignal = newExponentialSmoothing(10),
    airspeed = newExponentialSmoothing(10),
    altitude = newExponentialSmoothing(10)
}

local lightsState = 0

local wheelObj = -1
local wheelRadius = 0

local signalRightState = false
local signalLeftState = false
local signalWarnState = false

local blinkPulse = false
local blinkTimerThreshold = 0.4
local blinkTimer = 0
local rpmspin = 0

local function generateBlinkPulse(dt)
    blinkTimer = blinkTimer + dt
    if blinkTimer > blinkTimerThreshold then
        blinkPulse = not blinkPulse
        blinkTimer = 0
        if signalRightState or signalLeftState or signalWarnState then
            if MFM.blinkerSound then
                sounds.playSoundOnceAtNode("ShiftTestSound", 0, 1)
            end
        end
    end
end

local function update(dt)
    generateBlinkPulse(dt)
    
    local vals = M.values
    -- the primary source values
    vals.running = (drivetrain.engineDisabled == false)

    -- steering
    local _, h
    for _, h in pairs (hydros.hydros) do
        if h.inputSource == "steering_input" then
            if h.hydroDirState then
                vals.steering = -h.hydroDirState * v.data.input.steeringWheelLock
                break
            end
        end
    end

    if wheelObj ~= nil then
        vals.wheelspeed = math.abs(drivetrain.avgAV) * wheelRadius
    end

    vals.driveshaft = drivetrain.driveshaftAngle * (360 / math.pi)
    vals.axle = drivetrain.axleAngle * (360 / math.pi)    

    -- P R N D 2 1
    -- 0 1 2 3 4 5
    if drivetrain.shifterMode == 3 then
        vals.gear_A = drivetrain.shifterPosition + 2
    else
        vals.gear_A = math.min(math.max(drivetrain.gear + 2, 1), 3)
        if vals.parkingbrake_input > 0 then vals.gear_A = 0 end
    end
    vals.gear_A = vals.gear_A * (1/5)

    vals.gear_M = drivetrain.gear

    if obj.wheel_count > 0 then
        vals.rpm = drivetrain.rpm
        rpmspin = rpmspin + (dt * drivetrain.rpm)
        if rpmspin > 360 then rpmspin = rpmspin - 360 end
        vals.rpmspin = rpmspin
    end

    vals.brake = drivetrain.brake
    vals.throttle = drivetrain.throttle    
    vals.clutch = drivetrain.clutch    
    vals.parkingbrake = vals.parkingbrake_input
    vals.lights = lightsState    
    vals.fuel = drivetrain.fuel / drivetrain.fuelCapacity    
    vals.oiltemp = 0.3    
    vals.watertemp = 0.4    
    if signalWarnState then vals.turnsignal = 0
    elseif signalRightState then vals.turnsignal = 1
    elseif signalLeftState then vals.turnsignal = -1
    else vals.turnsignal = 0
    end    
    vals.airspeed = math.abs(obj:getVelocity():length())    
    vals.altitude = obj:getPosition().z    
    vals.parking = 0 -- TODO: input.parkinglights    
    vals.reverse = (drivetrain.gear < 0)

    -- highbeam
    
    -- and then the derived values
    vals.signal_L = (vals.signal_left_input == 1 and blinkPulse)
    vals.signal_R = (vals.signal_right_input == 1 and blinkPulse)
    vals.signal_right_input = (signalRightState)
    vals.signal_left_input = (signalLeftState)    
    vals.hazard = (signalWarnState and blinkPulse)
    vals.checkengine = (drivetrain.engineDisabled == true)
    vals.ignition = (drivetrain.engineDisabled == false)
    vals.lowfuel = (drivetrain.fuel < (0.1 * drivetrain.fuelCapacity))
    vals.lowpressure = (beamstate.lowpressure)
    vals.abs = drivetrain.absPulse    
    vals.esc = drivetrain.esc.escPulse    
    vals.lowhighbeam = (lightsState == 1 or lightsState == 2)
    vals.lowbeam = (lightsState == 1)
    vals.highbeam = (lightsState == 2)

    for f, v in pairs(vals) do
        if M.disabledState[f] ~= nil then
            vals[f] = nil
        else
            if type(v) == 'boolean' then
                 vals[f] = vals[f] and 1 or 0
            end
        end
    end

    for f, s in pairs(smoothers) do
        if vals[f] ~= nil then
            vals[f] = s:get(vals[f])
        end
    end

    for f, v in pairs(vals) do
        if v ~= lastValues[f] then
            material.funcChanged(f, v)
        end
        lastValues[f] = v
    end

    gui.send("electrics", vals)
end

local function reset()
    M.disabledState = {}

    -- find wheel to get wheelspeed from
    wheelObj = nil
    wheelRadius = 0
    for wi,wd in pairs(drivetrain.wheelInfo) do
        if wd.speedo ~= nil and wd.speedo == true then
            wheelObj = obj:getWheel(wd.wheelID)
            wheelRadius = wd.radius
            break
        end
    end
end

local function beamBroke(id)
--[[
    if not v.data.props then return end
    local brokenBeamName = v.data.beams[id].name
    if not brokenBeamName then
        return
    end

    for propKey, prop in pairs (v.data.props) do
        if prop.disableBeams then
            for k,ab in pairs(prop.disableBeams) do
                if ab == brokenBeamName then
                    log('D', "electrics.beamBroke", "prop beam broke:"..v)
                    prop.disabled = 1
                    break
                end
            end
        end
    end
]]--
end

-- user input functions
local function toggle_left_signal()
    if not signalWarnState then
        signalLeftState = not signalLeftState
    else
        signalLeftState = true
    end
    if signalLeftState then
        signalRightState = false
        signalWarnState = false
    end
end

local function toggle_right_signal()
    if not signalWarnState then
        signalRightState = not signalRightState
    else
        signalRightState = true
    end    
    if signalRightState then
        signalLeftState = false
        signalWarnState = false
    end
end

local function toggle_warn_signal()
    signalWarnState = not signalWarnState
    signalRightState = signalWarnState
    signalLeftState  = signalWarnState
end

local function toggle_lights()
    lightsState = lightsState + 1
    if lightsState == 3 then lightsState = 0 end
end

local function setLightsState(newval)
    lightsState = newval
end

-- public interface
M.update = update
M.toggle_left_signal  = toggle_left_signal
M.toggle_right_signal = toggle_right_signal
M.toggle_warn_signal  = toggle_warn_signal
M.toggle_lights       = toggle_lights
M.setLightsState      = setLightsState
M.beamBroke           = beamBroke
M.reset               = reset
M.init                = reset

return M