-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

log = function(...) Lua:log(...) end
print = function(...) Lua:log('A', "print", tostring(...)) end

package.path = "lua/vehicle/?.lua;lua/vehicle/jbeam/?.lua;lua/common/?.lua;lua/?.lua;?.lua"
require("utils")
require('mathlib')

-- improve stacktraces
local STP = require "StackTracePlus"
debug.traceback = STP.stacktrace

perf = require("perf")
settings = require('simplesettings')

vehicleMode = 0
objectSlotID = 0

local engineShiftUpRPM = 0

local function nop() end

-- step functions
function physicsStep(dts)
    drivetrain.updateWheels(dts)
    sensors.update()
    thrusters.update()
    hydros.update(dts)
    ai.update(dts)
end

-- This is called in the global scope, so it is safe to do things that contact things outside the vehicle
function graphicsStep(dt)
    input.update(dt) -- must come first ;)
    drivetrain.updateEngine(dt)
    electrics.update(dt)
    props.update()
    beamstate.updateGFX()
    bullettime.update(dt)
    --outsim.updateGFX(dt)
    extensions.hook('updateGFX', dt)
    if vehicleMode == 1 then
        recovery.updateGFX(dt)
        outgauge.updateGFX(dt)
        guihooks.updateGFX()
        perf.update()

        -- void playRPMLeds(const float currentRPM, const float rpmFirstLedTurnsOn, const float rpmRedLine)
        --gameEngine:playRPMLeds(drivetrain.rpm, v.data.engine.idleRPM, v.data.engine.shiftUpRPM)

        gameEngine:playRPMLeds(drivetrain.rpm, engineShiftUpRPM * 0.8, engineShiftUpRPM)
        --print('playRPMLeds| currentRPM=' .. tostring(drivetrain.rpm) .. ' , rpmFirstLedTurnsOn=' .. tostring(v.data.engine.shiftUpRPM * 0.8) .. ' , rpmRedLine =' .. tostring(v.data.engine.shiftUpRPM))
    end

    --[[
    local rv = obj:getDirectionVector()
    log('D', "Vehicle:pushToPhysics", "direction: " .. rv.x .. ", " .. rv.y .. ", " .. rv.z)

    local r = obj:getRotation()
    log('D', "Vehicle:pushToPhysics", "rotation: " .. r.x .. ", " .. r.y .. ", " .. r.z .. ", " .. r.w)
    ]]
end

-- This is called in the local scope, so it is NOT safe to do things that contact things outside the vehicle
-- It is called right after graphicsStep
function graphicsStepLocal(dt)
    sounds.update(dt)
    ai.updateGFX(dt)
    hydros.updateGFX(dt)
    sensors.updateGFX(dt)
    thrusters.updateGFX()
    if vehicleMode == 1 then
        streams.updateGFX()
        gui.updateGFX(dt)
    end
    drivetrain.graphicsStepLocal()
    mapmgr.updateGFX()
    scenario.updateGFX()
    mapmgr.sendTracking()
end

-- debug rendering
function debugDraw(x, y, z)
    local focusPos = float3(0,0,0)
    if x and y and z then focusPos = float3(x,y,z) end
    bdebug.debugDraw(focusPos)
    ai.debugDraw(focusPos)
end

function enableCallbacks()
    -- this enables the most important frame callbacks
    -- in case of an error, they still point to nop and thus do not spam the console
    -- TODO: FIXME
end

function initSystems()
    objectSlotID = obj:getSlotID()
    material.init()
    drivetrain.init()
    sensors.reset()
    beamstate.init()
    thrusters.init()
    hydros.init()
    input.init()
    sounds.init()
    props.init()
    electrics.init()
    outgauge.init()
    --outsim.init()
    MFM.init()
    recovery.init()
    bdebug.init()
    scenario.init()
    sensors.init()

    extensions.hook('onInit')
end

function init(path, partConfigData)
    if not obj then log('W', "default.init", "Error getting main object: unable to spawn") ; return end
    log('D', "default.init", "spawning vehicle "..tostring(path)..' with parts: '..tostring(partConfigData) )
    -- we change the lookup path here, so it prefers the vehicle lua
    package.path = path .. "/lua/?.lua;".. package.path
    extensions = require("extensions")
    extensions.loadModules(path.."/lua")
    drivetrain = require("drivetrain")
    MFM = require("MFM")
    sounds     = require("sounds")
    bdebug     = require("bdebug")
    input      = require("input")
    props      = require("props")

    particlefilter = require("particlefilter")
    particles  = require("particles")
    material   = require("material")
    v          = require("jbeam_main")
    electrics  = require("electrics")
    json       = require("json")
    beamstate  = require("beamstate")
    sensors    = require("sensors")
    bullettime = require("bullettime")
    thrusters  = require("thrusters")
    hydros     = require("hydros")
    gui        = require("gui") -- do not change its name, the GUI callback will break otherwise
    partmgmt   = require("partmgmt") -- do not change its name, the GUI callback will break otherwise
    streams    = require("guistreams")
    guihooks   = require("guihooks")
    outgauge   = require("outgauge")
    ai         = require("ai")
    --outsim     = require("outsim")
    recovery   = require("recovery")
    mapmgr     = require("mapmgr")
    scenario   = require("scenario")

    objectSlotID = obj:getSlotID()
    local hp = HighPerfTimer()

    -- care about the config before pushing to the physics
    if type(partConfigData) == 'string' and string.len(partConfigData) > 0 then
        local firstChar = string.sub(partConfigData,1,1)
        if firstChar == '{' or firstChar == '[' then
            log('D', "default.init", "using partconfig data "..tostring(partConfigData)..' - path:'..tostring(path) )
            partmgmt.setConfig(unserialize(partConfigData), false)
        else
            log('D', "default.init", "using partconfig filename "..tostring(partConfigData)..' - path:'..tostring(path) )
            partmgmt.load(partConfigData, false)
        end
    end

    -- load jbeam file
    if not v.loadDirectory(path) then
        return
    end

    --Settings.spawnScale = 1.6

    -- you can change the data in here before it gets submitted to the physics

    -- submit to physics
    v.pushToPhysics(obj)

    if v.data == nil then
        v.data = {}
    end

    initSystems()

    -- temporary camera settings
    obj.camRelaxation = 3

    -- temporary tire mark setting
    obj.slipTireMarkThreshold = 15

    -- if we got to here, lets enable the callbacks and hope for the best
    enableCallbacks()

    if v.data.engine ~= nil then
        engineShiftUpRPM = v.data.engine.shiftUpRPM or 0
    end

    log('D', "default.init", "init done - vehicle loading took "..hp:stop()..' ms')
end

-- various callbacks
function beamBroken(id, energy)
    beamstate.beamBroken(id, energy)
    --bullettime.slowMotion(0.05, 20) -- 20% of realtime
end

function beamDeformed(id, ratio)
    beamstate.beamDeformed(id, ratio)
end

-- called when everything was cleared
function vehicleDestroy()
    --log('D', "default.vehicleDestroy", "vehicleDestroy()")
    -- when the vehicle gets unloaded, remove all sounds
    sounds.destroy()
    material.destroy()
end

-- called when the user pressed I
function vehicleResetted(retainDebug)
    --log('D', "default.vehicleResetted", "vehicleResetted()")
    objectSlotID = obj:getSlotID()
    drivetrain.reset()
    props.reset()
    sensors.reset()
    if not retainDebug then
        bdebug.reset()
    end
    bullettime.reset()
    beamstate.reset()
    thrusters.reset()
    input.reset()
    hydros.reset()
    material.reset()
    electrics.reset()
    guihooks.reset()
    extensions.hook('onReset', retainDebug)
    ai.reset()
    partmgmt.vehicleResetted()
    mapmgr.reset()
    scenario.reset()
end

function particleEmitted(p)
    drivetrain.updateWheelSlip(p)
    particlefilter.particleEmitted(p)
end

function activated(mode)
    vehicleMode = mode
    objectSlotID = obj:getSlotID()
    if vehicleMode == 1 then mapmgr.enableTracking() end
    if vehicleMode == 0 then mapmgr.disableTracking() end
    --[[
    if mode == 1 then
        log('D', "default.activated", "yay, we got active")
    else
        log('D', "default.activated", "deactivated :(")
    end
    ]]--

    bdebug.activated(mode)
    extensions.hook('activated', mode)
    guihooks.trigger("VehicleFocusChanged", {slotid = objectSlotID, mode = mode})
end

function instabilityDetected()
    -- enable break debug, so we see what beams broke
    local msg = "instability detected for vehicle " .. tostring(v.vehicleDirectory) .. ": resetting the vehicle and disabling the physics."
    log('E', "default.instabilityDetected", msg)

    --obj:executeJS("HookManager.trigger('Message', "..encodeJson({msg=msg, ttl=10, category='instability'})..");")

    BeamEngine:setEnabled(false)
    obj:requestReset(RESET_PHYSICS)

   --if settings and settings.getValue('disableDynamicCollision') == true then
   --  Settings.dynamicCollision = false
   --end
end

function exportPersistentData()
    local s = serializeModules()
    --log('D', "default.exportPersistentData", s)
    obj:setPersistentData(s)
end

function importPersistentData(s)
    --log('D', "default.importPersistentData", s)
    deserializeModules(unserialize(s))
end

function guiUpdate(d)
    --log('D', "default.guiUpdate", "guiUpdate()")
    gui.gotUpdate(d)
end

